import casadi as ca
import numpy as np

from MotionPlanningSceneHelpers.referenceTrajectory import ReferenceTrajectory, t


class TrajectoryComponentMissingError(Exception):
    pass


class AnalyticTrajectory(ReferenceTrajectory):
    def __init__(self, n: int, **kwargs):
        super().__init__(n)
        if 'traj' not in kwargs:
            raise TrajectoryComponentMissingError("Trajectory definition not complete. Missing component: traj")
        self._t = t
        self._traj = ca.vcat([eval(x) for x in kwargs.get('traj')])

    def concretize(self):
        self._v = ca.jacobian(self._traj, self._t)
        self._a = ca.jacobian(self._v, self._t)
        self._funs = ca.Function("traj", [self._t], [self._traj, self._v, self._a])

    def shuffle(self):
        pass

    def evaluate(self, t):
        fun = self._funs(t)
        x = np.array(fun[0])[:, 0]
        v = np.array(fun[1])[:, 0]
        a = np.array(fun[2])[:, 0]
        return [x, v, a]



from dataclasses import dataclass
from typing import Optional, Any
import os
import numpy as np
import csv
from omegaconf import OmegaConf

from MotionPlanningSceneHelpers.analyticTrajectory import AnalyticTrajectory
from MotionPlanningSceneHelpers.splineTrajectory import SplineTrajectory
from MotionPlanningSceneHelpers.motionPlanningComponent import (
    ComponentIncompleteError,
    DimensionNotSuitableForEnv,
)
from MotionPlanningEnv.collisionObstacle import (
    CollisionObstacle,
    CollisionObstacleConfig,
)


class DynamicSphereObstacleMissmatchDimensionError(Exception):
    pass


class TypeNotSupportedError(Exception):
    pass


@dataclass
class GeometryConfig:
    """Configuration dataclass for geometry.

    This configuration class holds information about position
    and radius of a dynamic sphere obstacle.

    Parameters:
    ------------

    trajectory: Any
        trajectory description of the obstacle. Can be either a spline or
        a analytic trajectory.
    radius: float
        radius of the obstacle
    """

    trajectory: Any
    radius: float


@dataclass
class DynamicSphereObstacleConfig(CollisionObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a dynamic spherical obstacle.

    Parameters:
    ------------

    geometry : GeometryConfig : Geometry of the obstacle
    low: GeometryConfig : Lower limit for randomization
    high: GeometryConfig : Upper limit for randomization
    """

    geometry: GeometryConfig
    low: Optional[GeometryConfig] = None
    high: Optional[GeometryConfig] = None


class DynamicSphereObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        schema = OmegaConf.structured(DynamicSphereObstacleConfig)
        super().__init__(schema, **kwargs)
        self.check_completeness()
        self.check_dimensionality()
        if self.type() == "splineSphere":
            self._traj = SplineTrajectory(
                self.dimension(), traj=self._config.geometry.trajectory
            )
        elif self.type() == "sphere" or self.type() == "analyticSphereObstacle":
            self._traj = AnalyticTrajectory(
                self.dimension(), traj=self._config.geometry.trajectory
            )
        elif self.type() == "analyticSphere":
            self._traj = AnalyticTrajectory(
                self.dimension(), traj=self._config.geometry.trajectory
            )
        self._traj.concretize()

    def check_dimensionality(self):
        if self.type() == "splineSphere":
            dim_verification = len(
                self.geometry()["trajectory"]["controlPoints"][0]
            )
        elif self.type() == "sphere" or self.type() == "analyticSphereObstacle":
            dim_verification = len(self.geometry()["trajectory"])
        elif self.type() == "analyticSphere":
            dim_verification = len(self.geometry()["trajectory"])
        else:
            raise TypeNotSupportedError(
                f"Obstacle type {self.type()} not supported"
            )
        if self.dimension() != dim_verification:
            raise DynamicSphereObstacleMissmatchDimensionError(
                "Dimension mismatch between trajectory array and dimension"
            )

    def dimension(self):
        if self.type() in ["sphere", "analyticSphere"]:
            return len(self._config.geometry.trajectory)
        elif self.type() in ["splineSphere"]:
            return len(self._config.geometry.trajectory.controlPoints[0])

    def traj(self):
        return self._traj

    def check_geometry_completeness(self):
        incomplete = False
        missing_keys = ""
        for key in self._geometry_keys:
            if key not in self.geometry():
                incomplete = True
                missing_keys += key + ", "
        if incomplete:
            raise ComponentIncompleteError(
                f"Missing keys in geometry: {missing_keys[:-2]}"
            )

    def position(self, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        return self._traj.evaluate(t)[0]

    def velocity(self, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        return self._traj.evaluate(t)[1]

    def acceleration(self, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        return self._traj.evaluate(t)[2]

    def radius(self):
        return self.geometry()["radius"]


    def movable(self):
        return False

    def csv(self, file_name, samples=100):
        theta = np.arange(-np.pi, np.pi, step=np.pi / samples)
        x = self.position()[0] + (self.radius() - 0.1) * np.cos(theta)
        y = self.position()[1] + (self.radius() - 0.1) * np.sin(theta)
        with open(file_name, mode="w") as file:
            csv_writer = csv.writer(file, delimiter=",")
            for i in range(2 * samples):
                csv_writer.writerow([x[i], y[i]])

    def render_gym(self, viewer, rendering, **kwargs):
        x = self.position(t=kwargs.get("t"))
        if self.dimension() != 2:
            raise DimensionNotSuitableForEnv(
                "PlanarGym only supports two dimensional obstacles"
            )
        tf = rendering.Transform(rotation=0, translation=(x[0], x[1]))
        joint = viewer.draw_circle(self.radius())
        joint.add_attr(tf)

    def add_to_bullet(self, pybullet) -> int:
        if self.dimension() == 2:
            base_position= self.position().tolist() + [0.0]
        elif self.dimension() == 3:
            base_position= self.position().tolist()
        else:
            raise DimensionNotSuitableForEnv(
                "Pybullet only supports three dimensional obstacles"
            )
        collision_shape = pybullet.createCollisionShape(
            pybullet.GEOM_SPHERE,
            radius=self.radius(),
        )
        base_orientation = [0, 0, 0, 1]
        mass = int(self.movable())

        pybullet.setAdditionalSearchPath(
            os.path.dirname(os.path.realpath(__file__))
        )
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_MESH,
            fileName="sphere_smooth.obj",
            rgbaColor=[1.0, 0.0, 0.0, 1.0],
            specularColor=[1.0, 0.5, 0.5],
            meshScale=[self.radius(), self.radius(), self.radius()],
        )

        assert isinstance(base_position, list)
        assert isinstance(base_orientation, list)
        self._bullet_id = pybullet.createMultiBody(
            mass, collision_shape, visual_shape_id, base_position, base_orientation
        )
        return self._bullet_id

    def update_bullet_position(self, pybullet, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        pos = self.position(t=t).tolist()
        if self.dimension() == 2:
            pos += [0.0]
        ori = [0, 0, 0, 1]
        pybullet.resetBasePositionAndOrientation(self._bullet_id, pos, ori)



from dataclasses import dataclass
from typing import List, Optional, Any
import numpy as np
from omegaconf import OmegaConf
from pyquaternion import Quaternion
from MotionPlanningGoal.subGoal import SubGoal, SubGoalConfig
from MotionPlanningSceneHelpers.motionPlanningComponent import (
    DimensionNotSuitableForEnv,
)
from MotionPlanningSceneHelpers.analyticTrajectory import AnalyticTrajectory
from MotionPlanningSceneHelpers.splineTrajectory import SplineTrajectory


class TrajectoryTypeNotSupported(Exception):
    pass


@dataclass
class DynamicSubGoalConfig(SubGoalConfig):
    """Configuration dataclass for static sub goal.

    This configuration class holds information about the
    the weight, accuracy required, type and position in the
    kinematic chain.

    Parameters:
    ------------

    parent_link: str
        Name of the link that specifies the frame in which the goal is defined
    child_link: str
        Name of the link that should match the desired position
    trajectory: Any
        Trajectory of the goal, either defined by spline or analytic equation.
    angle list
        Additional rotation from the parent_link frame given by a quaternion
    low : list
        Lower limit for randomization
    high : list
        Upper limit for randomization

    """

    parent_link: str
    child_link: str
    trajectory: Any
    angle: Optional[Any] = None
    low: Optional[List[float]] = None
    high: Optional[List[float]] = None


class DynamicSubGoal(SubGoal):
    def __init__(self, **kwargs):
        schema = OmegaConf.structured(DynamicSubGoalConfig)
        super().__init__(schema, **kwargs)
        self.check_completeness()
        if self.type() == "splineSubGoal":
            self._traj = SplineTrajectory(
                self.dimension(), traj=self._config.trajectory
            )
        elif self.type() == "analyticSubGoal":
            self._traj = AnalyticTrajectory(
                self.dimension(), traj=self._config.trajectory
            )
        else:
            raise TrajectoryTypeNotSupported(
                f"Trajectory type {self.type()} not supported"
            )
        self._traj.concretize()
        self.check_dimensionality()

    def parent_link(self):
        return self._config.parent_link

    def child_link(self):
        return self._config.child_link

    def traj(self):
        return self._traj

    def dict(self):
        return OmegaConf.to_container(self._config)

    def position(self, **kwargs) -> np.ndarray:
        return self.evaluate_trajectory(**kwargs)[0]

    def velocity(self, **kwargs) -> np.ndarray:
        return self.evaluate_trajectory(**kwargs)[1]

    def acceleration(self, **kwargs) -> np.ndarray:
        return self.evaluate_trajectory(**kwargs)[2]

    def evaluate_trajectory(self, **kwargs):
        t = kwargs.get("t") if "t" in kwargs else 0.0
        return self._traj.evaluate(t=t)

    def shuffle(self):
        self._traj.shuffle()

    def angle(self):
        return self._config.angle

    def render_gym(self, viewer, rendering, **kwargs):
        coordinate_system_1 = viewer.draw_line(
            [-3, 0], [3, 0], color=[0.0, 0.0, 0.0]
        )
        coordinate_system_2 = viewer.draw_line(
            [0, -3], [0, 3], color=[0.0, 0.0, 0.0]
        )
        angle = self.angle()
        if not angle:
            angle = 0.0
        tf2 = rendering.Transform(rotation=angle)
        coordinate_system_1.add_attr(tf2)
        coordinate_system_2.add_attr(tf2)
        if self.dimension() == 1:
            start_point = [-10, -10.0]
            end_point = [10.0, 10.0]
            start_point[self.indices()[0]] = 0.0
            end_point[self.indices()[0]] = 0.0
            goal = viewer.draw_line(
                start_point, end_point, color=[0.0, 1.0, 0.0]
            )
            translation = [0.0, 0.0]
            translation[self.indices()[0]] = self.position(t=kwargs.get("t"))[0]
            tf = rendering.Transform(translation=translation)
            goal.add_attr(tf)
            goal.add_attr(tf2)
        elif self.dimension() == 2:
            x = self.position(t=kwargs.get("t"))
            goal = viewer.draw_circle(self.epsilon(), color=[0.0, 1.0, 0.0])
            tf = rendering.Transform(translation=(x[0], x[1]))
            goal.add_attr(tf)
            goal.add_attr(tf2)
        else:
            raise DimensionNotSuitableForEnv(
                "PlanarGym only supports two dimensional obstacles"
            )

    def update_bullet_position(self, pybullet, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        pos = self.position(t=t).tolist()
        if self.dimension() == 2 and self.indices() == [0, 1]:
            pos += [0.0]
        ori = [0, 0, 0, 1]
        pybullet.resetBasePositionAndOrientation(self._bullet_id, pos, ori)

    def add_to_bullet(self, pybullet, position=[0.0, 0.0, 0.0]) -> int:
        if self.dimension() == 2:
            base_position = self.position().tolist() + [0.0]
        elif self.dimension() == 3:
            base_position = self.position().tolist()
        else:
            raise DimensionNotSuitableForEnv(
                "Pybullet only supports three dimensional obstacles"
            )
        rgba_color = [0.0, 1.0, 0.0, 0.3]
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_SPHERE, rgbaColor=rgba_color, radius=self.epsilon()
        )
        collision_shape = -1
        base_orientation = [0, 0, 0, 1]
        mass = 0

        assert isinstance(base_position, list)
        assert isinstance(base_orientation, list)
        self._bullet_id = pybullet.createMultiBody(
            mass,
            collision_shape,
            visual_shape_id,
            base_position,
            base_orientation,
        )

        if self.angle():
            for i in range(3):
                self.add_axis_component(pybullet, i, position)
        return self._bullet_id


    def add_axis_component(self, pybullet, i, goal_position):
        rgba_color = [0.0, 0.0, 0.0, 0.3]
        angles = self.angle()
        if not angles:
            angles = Quaternion([1, 0.0, 0, 0])
        else:
            angles = Quaternion(angles).inverse
        orientation = Quaternion([1, 0, 0, 0])
        orientation[i + 1] = 1.0
        orientation = orientation.normalised
        position = np.array(
            [goal_position[0], goal_position[1], goal_position[2]]
        )
        orientation = angles * orientation
        orientation_array = [
            orientation[1],
            orientation[2],
            orientation[3],
            orientation[0],
        ]
        offset_array = [
            np.array([0.00, 0.00, -0.05]),
            np.array([0.0, 0.0, 0.05]),
            np.array([0.0, 0.0, 0.05]),
        ]

        position += orientation.rotate(offset_array[i])
        index_map = [1, 0, 2]
        rgba_color[index_map[i]] = 1.0
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_CYLINDER,
            rgbaColor=rgba_color,
            radius=0.01,
            length=0.1,
        )
        collision_shape = -1
        mass = 0
        assert isinstance(position, list)
        assert isinstance(orientation_array, list)
        return pybullet.createMultiBody(
            mass,
            collision_shape,
            visual_shape_id,
            position,
            orientation_array,
        )



from MotionPlanningSceneHelpers.motionPlanningComponent import (
    MotionPlanningComponent,
)
from MotionPlanningGoal.subGoal import SubGoal
from MotionPlanningGoal.subGoalCreator import SubGoalCreator
from MotionPlanningGoal.staticJointSpaceSubGoal import (
    JointSpaceGoalsNotSupportedError,
)

import yaml
import logging
from omegaconf import OmegaConf


class MultiplePrimeGoalsError(Exception):
    pass


class GoalComposition(MotionPlanningComponent):
    def __init__(self, **kwargs):
        if "content_dict" in kwargs and "name" in kwargs:
            self._content_dict = kwargs.get("content_dict")
            self._name = kwargs.get("name")
        elif "file_name" in kwargs:
            with open(kwargs.get("file_name"), "r") as stream:
                self._content_dict = yaml.safe_load(stream)
            self._name = self._content_dict["name"]
            del self._content_dict["name"]
        self._config = OmegaConf.create(self._content_dict)
        self._primary_goal_index = -1
        self._sub_goals = []
        self._sub_goal_creator = SubGoalCreator()
        self.parse_sub_goals()

    def parse_sub_goals(self):
        for sub_goal_name in self._config.keys():
            sub_goal_type = self._config[sub_goal_name].type
            sub_goal_dict = self._config[sub_goal_name]
            sub_goal = self._sub_goal_creator.create_sub_goal(
                sub_goal_type, sub_goal_name, sub_goal_dict
            )
            if sub_goal.is_primary_goal():
                if self._primary_goal_index >= 0:
                    raise MultiplePrimeGoalsError(
                        "There are multiple prime goals. Using first prime goal"
                    )
                else:
                    self._primary_goal_index = len(self._sub_goals)
            self._sub_goals.append(sub_goal)

    def primary_goal(self):
        return self.get_goal_by_index(self._primary_goal_index)

    def sub_goals(self):
        return self._sub_goals

    def get_goal_by_name(self, name) -> SubGoal:
        for sub_goal in self._sub_goals:
            if sub_goal.name() == name:
                return sub_goal

    def get_goal_by_index(self, index):
        return self._sub_goals[index]

    def evaluate(self, **kwargs):
        evals = []
        for sub_goal in self._sub_goals:
            evals += sub_goal.evaluate(**kwargs)
        return evals

    def dict(self):
        composition_dict = {}
        for sub_goal in self._sub_goals:
            composition_dict[sub_goal.name()] = sub_goal.dict()
        return composition_dict

    def shuffle(self):
        for sub_goal in self._sub_goals:
            sub_goal.shuffle()

    def render_gym(self, viewer, rendering, **kwargs):
        for sub_goal in self._sub_goals:
            try:
                sub_goal.render_gym(viewer, rendering, **kwargs)
            except JointSpaceGoalsNotSupportedError as _:
                logging.warn("Skipping visualization of joint space goal.")

    def add_to_bullet(self, pybullet):
        for sub_goal in self._sub_goals:
            try:
                sub_goal.add_to_bullet(
                    pybullet, position=self.primary_goal().position()
                )
            except JointSpaceGoalsNotSupportedError as _:
                print("Skipping visualization of joint space goal.")

    def update_bullet_position(self, pybullet, **kwargs):
        self.primary_goal().update_bullet_position(pybullet, **kwargs)



import numpy as np
from MotionPlanningGoal.subGoal import SubGoal, SubGoalConfig
from MotionPlanningSceneHelpers.motionPlanningComponent import (
    DimensionNotSuitableForEnv,
)
from pyquaternion import Quaternion

from omegaconf import OmegaConf
from typing import List, Optional, Any
from dataclasses import dataclass


@dataclass
class StaticSubGoalConfig(SubGoalConfig):
    """Configuration dataclass for static sub goal.

    This configuration class holds information about the
    the weight, accuracy required, type and position in the
    kinematic chain.

    Parameters:
    ------------

    parent_link: str
        Name of the link that specifies the frame in which the goal is defined
    child_link: str
        Name of the link that should match the desired position
    desired_position : list
        Goal state of the concerned link
    angle: list
        Additional rotation from the parent_link frame given by a quaternion
    low: list
        Lower limit for randomization
    high: list
        Upper limit for randomization

    """

    parent_link: Any
    child_link: Any
    desired_position: List[float]
    angle: Optional[Any] = None
    low: Optional[List[float]] = None
    high: Optional[List[float]] = None


class StaticSubGoal(SubGoal):
    def __init__(self, **kwargs):
        schema = OmegaConf.structured(StaticSubGoalConfig)
        super().__init__(schema, **kwargs)
        self.check_completeness()
        self.check_dimensionality()

    def parent_link(self):
        return self._config.parent_link

    def child_link(self):
        return self._config.child_link

    def limit_low(self):
        if self._config.low:
            return np.array(self._config.low)
        else:
            return np.ones(self.dimension()) * -1

    def limit_high(self):
        if self._config.high:
            return np.array(self._config.high)
        else:
            return np.ones(self.dimension()) * 1

    def position(self, **kwargs) -> np.ndarray:
        return np.array(self._config.desired_position)

    def velocity(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def acceleration(self, **kwargs):
        return np.zeros(self.dimension())

    def shuffle(self):
        random_pos = np.random.uniform(
            self.limit_low(), self.limit_high(), self.dimension()
        )
        self._config.desired_position = random_pos.tolist()

    def angle(self):
        if isinstance(self._config.angle, float):
            return self._config.angle
        if self._config.angle:
            return list(self._config.angle)

    def render_gym(self, viewer, rendering, **kwargs):
        """Rendering the static sub goal into a planar gym environment.

        The static sub goal is rendered into the viewer of the open-ai gym by
        adding either a point or a line in the rendering window.

        Parameters:
        ------------
        viewer: Viewing window of open-ai gym
        rendering: Rendering toolbox in open-ai gym
        """
        coordinate_system_1 = viewer.draw_line(
            [-3, 0], [3, 0], color=[0.0, 0.0, 0.0]
        )
        coordinate_system_2 = viewer.draw_line(
            [0, -3], [0, 3], color=[0.0, 0.0, 0.0]
        )
        angle = self.angle()
        if angle:
            tf2 = rendering.Transform(rotation=self.angle())
        else:
            tf2 = rendering.Transform()
        coordinate_system_1.add_attr(tf2)
        coordinate_system_2.add_attr(tf2)
        if self.dimension() == 1:
            start_point = [-10, -10.0]
            end_point = [10.0, 10.0]
            start_point[self.indices()[0]] = 0.0
            end_point[self.indices()[0]] = 0.0
            goal = viewer.draw_line(
                start_point, end_point, color=[0.0, 1.0, 0.0]
            )
            translation = [0.0, 0.0]
            translation[self.indices()[0]] = self.position()[0]
            tf = rendering.Transform(translation=translation)
            goal.add_attr(tf)
            goal.add_attr(tf2)
        elif self.dimension() == 2:
            x = self.position()
            goal = viewer.draw_circle(self.epsilon(), color=[0.0, 1.0, 0.0])
            tf = rendering.Transform(translation=(x[0], x[1]))
            goal.add_attr(tf)
            goal.add_attr(tf2)
        else:
            raise DimensionNotSuitableForEnv(
                "PlanarGym only supports two dimensional obstacles"
            )

    def add_to_bullet(self, pybullet, position=[0.0, 0.0, 0.0]) -> int:
        """Adds the static sub goal to pybullet.

        The static sub goal is added to the pybullet as point or a coordinate
        system if the orientation is specified.

        Parameters:
        ------------
        pybullet: pybulle-instance
        position: list:
            goal position is default the origin for orientation-only goals.
        """
        if self.dimension() == 2:
            base_position = self.position().tolist() + [0.0]
        elif self.dimension() == 3:
            base_position = self.position().tolist()
        else:
            raise DimensionNotSuitableForEnv(
                "Pybullet only supports three dimensional obstacles"
            )
        rgba_color = [0.0, 1.0, 0.0, 0.3]
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_SPHERE, rgbaColor=rgba_color, radius=self.epsilon()
        )
        collision_shape = -1
        base_orientation = [0, 0, 0, 1]
        mass = 0

        assert isinstance(base_position, list)
        assert isinstance(base_orientation, list)
        multi_body_id = pybullet.createMultiBody(
            mass,
            collision_shape,
            visual_shape_id,
            base_position,
            base_orientation,
        )

        if self.angle():
            for i in range(3):
                self.add_axis_component(pybullet, i, position)
        return multi_body_id

    def add_axis_component(self, pybullet, i, goal_position):
        """Adds the coordinate system of the static sub goal to pybullet.

        If the user has specified an angle for the subgoal, the coordinate
        system is added to pybullet as three differently colored cylinders.

        Parameters:
        ------------
        pybullet: yybullet-instance
        i: int
            Axis specifier for either x(i=0), y(i=1), z(i=2)
        goal_position: list
            goal position to align with a potential position goal
        """
        rgba_color = [0.0, 0.0, 0.0, 0.3]
        angles = self.angle()
        if not angles:
            angles = Quaternion([1, 0.0, 0, 0])
        else:
            angles = Quaternion(angles).inverse
        orientation = Quaternion([1, 0, 0, 0])
        orientation[i + 1] = 1.0
        orientation = orientation.normalised
        position = np.array(
            [goal_position[0], goal_position[1], goal_position[2]]
        )
        orientation = angles * orientation
        orientation_array = [
            orientation[1],
            orientation[2],
            orientation[3],
            orientation[0],
        ]
        offset_array = [
            np.array([0.00, 0.00, -0.05]),
            np.array([0.0, 0.0, 0.05]),
            np.array([0.0, 0.0, 0.05]),
        ]

        position += orientation.rotate(offset_array[i])
        index_map = [1, 0, 2]
        rgba_color[index_map[i]] = 1.0
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_CYLINDER,
            rgbaColor=rgba_color,
            radius=0.01,
            length=0.1,
        )
        collision_shape = -1
        mass = 0
        pybullet.createMultiBody(
            mass,
            collision_shape,
            visual_shape_id,
            position,
            orientation_array,
        )



import numpy as np
from MotionPlanningGoal.subGoal import SubGoal, SubGoalConfig

from omegaconf import OmegaConf
from typing import List, Optional
from dataclasses import dataclass


class JointSpaceGoalsNotSupportedError(Exception):
    pass


@dataclass
class StaticJointSpaceSubGoalConfig(SubGoalConfig):
    """Configuration dataclass for static joint space sub goal.

    This configuration class holds information about the
    the desired joint configuration and the limits for randomization.

    Parameters:
    ------------

    desired_position : list : Goal configuration of the robot
    low : list : Lower limit for randomization
    high : list : Upper limit for randomization
    """

    desired_position: List[float]
    low: Optional[List[float]] = None
    high: Optional[List[float]] = None


class StaticJointSpaceSubGoal(SubGoal):
    def __init__(self, **kwargs):
        schema = OmegaConf.structured(StaticJointSpaceSubGoalConfig)
        super().__init__(schema, **kwargs)
        self.check_completeness()
        self.check_dimensionality()

    def limit_low(self):
        if self._config.low:
            return np.array(self._config.low)
        else:
            return np.ones(self.dimension()) * -1

    def limit_high(self):
        if self._config.high:
            return np.array(self._config.high)
        else:
            return np.ones(self.dimension()) * 1

    def evaluate(self, **kwargs):
        pass

    def position(self, **kwargs):
        return self._config.desired_position

    def velocity(self, **kwargs):
        return np.zeros(self.dimension())

    def acceleration(self, **kwargs):
        return np.zeros(self.dimension())

    def shuffle(self):
        random_pos = np.random.uniform(
            self.limit_low(), self.limit_high(), self.dimension()
        )
        self._config.desired_position = random_pos.tolist()

    def render_gym(self, viewer, **kwargs):
        raise JointSpaceGoalsNotSupportedError()

    def add_to_bullet(self, pybullet, **kwargs):
        raise JointSpaceGoalsNotSupportedError()



from MotionPlanningEnv.sphereObstacle import SphereObstacle
from MotionPlanningEnv.dynamicSphereObstacle import DynamicSphereObstacle


class UnknownObstacleType(Exception):
    pass


class ObstacleCreator(object):
    def __init__(self):
        pass

    def create_obstacle(self, obstacle_type, name, content_dict):
        if obstacle_type == "sphereObstacle":
            return SphereObstacle(name=name, content_dict=content_dict)
        elif obstacle_type == "analyticSphere":
            return DynamicSphereObstacle(name=name, content_dict=content_dict)
        elif obstacle_type == "splineSphere":
            return DynamicSphereObstacle(name=name, content_dict=content_dict)
        else:
            raise UnknownObstacleType(
                f"ObstacleType {obstacle_type} is not know"
            )



from dataclasses import dataclass
import numpy as np
import os
import csv
from MotionPlanningEnv.collisionObstacle import (
    CollisionObstacle,
    CollisionObstacleConfig,
)
from MotionPlanningSceneHelpers.motionPlanningComponent import (
    DimensionNotSuitableForEnv,
)

from omegaconf import OmegaConf
from typing import List, Optional


class SphereObstacleMissmatchDimensionError(Exception):
    pass


@dataclass
class GeometryConfig:
    """Configuration dataclass for geometry.

    This configuration class holds information about position
    and radius of a sphere obstacle.

    Parameters:
    ------------

    position: list: Position of the obstacle
    radius: float: Radius of the obstacle
    """

    position: List[float]
    radius: float


@dataclass
class SphereObstacleConfig(CollisionObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a spherical obstacle.

    Parameters:
    ------------

    geometry : GeometryConfig : Geometry of the obstacle
    movable : bool : Flag indicating whether an obstacle can be pushed around
    low : GeometryConfig : Lower limit for randomization
    high : GeometryConfig : Upper limit for randomization
    """

    geometry: GeometryConfig
    movable: bool = False
    low: Optional[GeometryConfig] = None
    high: Optional[GeometryConfig] = None


class SphereObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        schema = OmegaConf.structured(SphereObstacleConfig)
        super().__init__(schema, **kwargs)
        self.check_completeness()

    def dimension(self):
        return len(self._config.geometry.position)

    def limit_low(self):
        if self._config.low:
            return [
                np.array(self._config.low.position),
                self._config.low.radius,
            ]
        else:
            return [np.ones(self.dimension()) * -1, 0]

    def limit_high(self):
        if self._config.high:
            return [
                np.array(self._config.high.position),
                self._config.high.radius,
            ]
        else:
            return [np.ones(self.dimension()) * 1, 1]

    def position(self, **kwargs) -> np.ndarray:
        return np.array(self._config.geometry.position)

    def velocity(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def acceleration(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def radius(self):
        return self._config.geometry.radius

    def shuffle(self):
        random_pos = np.random.uniform(
            self.limit_low()[0], self.limit_high()[0], self.dimension()
        )
        random_radius = np.random.uniform(
            self.limit_low()[1], self.limit_high()[1], 1
        )
        self._config.geometry.position = random_pos.tolist()
        self._config.geometry.radius = float(random_radius)

    def movable(self):
        return self._config.movable

    def csv(self, file_name, samples=100):
        theta = np.arange(-np.pi, np.pi + np.pi / samples, step=np.pi / samples)
        x = self.position()[0] + (self.radius() - 0.1) * np.cos(theta)
        y = self.position()[1] + (self.radius() - 0.1) * np.sin(theta)
        with open(file_name, mode="w") as file:
            csv_writer = csv.writer(file, delimiter=",")
            for i in range(2 * samples):
                csv_writer.writerow([x[i], y[i]])

    def render_gym(self, viewer, rendering, **kwargs):
        if self.dimension() != 2:
            raise DimensionNotSuitableForEnv(
                "PlanarGym only supports two dimensional obstacles"
            )
        x = self.position()
        tf = rendering.Transform(rotation=0, translation=(x[0], x[1]))
        joint = viewer.draw_circle(self.radius())
        joint.add_attr(tf)

    def add_to_bullet(self, pybullet) -> int:
        if self.dimension() == 2:
            base_position = self.position().tolist() + [0.0]
        elif self.dimension() == 3:
            base_position = self.position().tolist()
        else:
            raise DimensionNotSuitableForEnv(
                "Pybullet only supports three dimensional obstacles"
            )
        collision_shape = pybullet.createCollisionShape(
            pybullet.GEOM_SPHERE, radius=self.radius()
        )
        visual_shape_id = -1
        base_orientation = [0, 0, 0, 1]
        mass = int(self.movable())
        pybullet.setAdditionalSearchPath(
            os.path.dirname(os.path.realpath(__file__))
        )
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_MESH,
            fileName="sphere_smooth.obj",
            rgbaColor=[1.0, 0.0, 0.0, 1.0],
            specularColor=[1.0, 0.5, 0.5],
            meshScale=[self.radius(), self.radius(), self.radius()],
        )
        assert isinstance(base_position, list)
        assert isinstance(base_orientation, list)
        return pybullet.createMultiBody(
            mass,
            collision_shape,
            visual_shape_id,
            base_position,
            base_orientation,
        )



import os

from MotionPlanningEnv.collisionObstacle import (
    CollisionObstacle,
    CollisionObstacleConfig,
)
from MotionPlanningSceneHelpers.motionPlanningComponent import (
    ComponentIncompleteError,
    DimensionNotSuitableForEnv,
)
import numpy as np

from dataclasses import dataclass
from omegaconf import OmegaConf
from typing import List, Optional


@dataclass
class GeometryConfig:
    """Configuration dataclass for geometry for urdf obstacles.

    This configuration class holds information about position.

    Parameters:
    ------------

    position: list: Position of the obstacle
    """

    position: List[float]


@dataclass
class UrdfObstacleConfig(CollisionObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a spherical obstacle.

    Parameters:
    ------------

    urdf : str : Filename of the urdf
    geometry : GeometryConfig : Geometry of the obstacle
    low: GeometryConfig : Lower limit for randomization
    high: GeometryConfig : Upper limit for randomization

    """

    urdf: str
    geometry: GeometryConfig
    low: Optional[GeometryConfig] = None
    high: Optional[GeometryConfig] = None


class UrdfObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        schema = OmegaConf.structured(UrdfObstacleConfig)
        super().__init__(schema, **kwargs)
        self.check_completeness()

    def urdf(self):
        return self._config.urdf

    def position(self):
        return self._config.geometry.position

    def velocity(self):
        return np.zeros(3)

    def acceleration(self):
        return np.zeros(3)

    def dimension(self):
        return len(self._config.geometry.position)

    def add_to_bullet(self, pybullet) -> int:
        if self.dimension() != 3:
            raise DimensionNotSuitableForEnv(
                "Pybullet only supports two dimensional obstacles"
            )
        return pybullet.loadURDF(fileName=self.urdf(), basePosition=self.position())



from abc import abstractmethod
from MotionPlanningSceneHelpers.motionPlanningComponent import MotionPlanningComponent

from dataclasses import dataclass

@dataclass
class CollisionObstacleConfig:
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the dimension and the type
    of collision obstacle.

    Parameters:
    ------------

    dimension : int : Dimension of the obstacle
    type : str : Type of the obstacle
    """
    type: str


class CollisionObstacle(MotionPlanningComponent):

    @abstractmethod
    def dimension(self):
        pass

    def type(self) -> str:
        return self._config.type

    def geometry(self):
        return self._config.geometry

    def evaluate(self, **kwargs) -> list:
        return [
            self.position(**kwargs),
            self.velocity(**kwargs),
            self.acceleration(**kwargs)
        ]

    @abstractmethod
    def position(self, **kwargs):
        pass

    @abstractmethod
    def velocity(self, **kwargs):
        pass

    @abstractmethod
    def acceleration(self, **kwargs):
        pass

    def update_bullet_position(self, pybullet, **kwargs):
        pass



from abc import ABC
import yaml
from omegaconf import OmegaConf


class ComponentIncompleteError(Exception):
    pass


class DimensionNotSuitableForEnv(Exception):
    pass


class MotionPlanningComponent(ABC):

    def __init__(self, schema, **kwargs):
        if 'content_dict' in kwargs and 'name' in kwargs:
            self._content_dict = kwargs.get('content_dict')
            self._name = kwargs.get('name')
        elif 'file_name' in kwargs:
            with open(kwargs.get('file_name'), 'r') as stream:
                self._content_dict = yaml.safe_load(stream)
            self._name = self._content_dict['name']
            del self._content_dict['name']
        self._config = OmegaConf.create(self._content_dict)
        config = OmegaConf.create(self._content_dict)
        self._config = OmegaConf.merge(schema, config)

    def check_completeness(self):
        pass
        """
        incomplete = False
        missingKeys = ""
        for key in self._required_keys:
            if key not in self._content_dict.keys():
                incomplete = True
                missingKeys += key + ", "
        if incomplete:
            raise ComponentIncompleteError("Missing keys: %s" % missingKeys[:-2])
        """

    def name(self):
        return self._name

    def evaluate_components(self, mask: list, t: float):
        result = {}
        for mask_entry in mask:
            try:
                result[mask_entry] = getattr(self, mask_entry)(t=t)
            except TypeError as _:
                result[mask_entry] = getattr(self, mask_entry)()
        return result

    def dict(self):
        return OmegaConf.to_container(self._config)

    def add_to_bullet(self, pybullet, position=[0.0, 0.0, 0.0]) -> int:
        """Adds component to pybullet instance.

        Returns
        ----------
        int: multi_body_identifier
        """
        pass



from abc import abstractmethod
from MotionPlanningSceneHelpers.motionPlanningComponent import (
    MotionPlanningComponent,
)
from dataclasses import dataclass

from typing import List


class SubGoalMissmatchDimensionError(Exception):
    pass


@dataclass
class SubGoalConfig:
    """Configuration dataclass for sub goal.

    This configuration class holds information about the
    the weight, accuracy required, type and position in the
    kinematic chain.

    Parameters:
    ------------

    m: int: Dimension of the sub goal
    w: float: Weight of the sub goal
    type: str: Type of the sub goal
    indices: list: Indices of a forward map to be considered
    epsilon: float: Required accuracy of the sub goal
    prime: bool: Flag for primary goal
    """

    weight: float
    type: str
    indices: List[int]
    epsilon: float
    is_primary_goal: bool


class SubGoal(MotionPlanningComponent):
    def check_dimensionality(self):
        if isinstance(self.position(), str):
            return
        if len(self.indices()) != len(self.position()):
            raise SubGoalMissmatchDimensionError(
                "Dimension mismatch between goal and indices"
            )

    def is_primary_goal(self):
        return self._config.is_primary_goal

    def epsilon(self):
        return self._config.epsilon

    def indices(self):
        return self._config.indices

    def dimension(self):
        return len(self.indices())

    def weight(self):
        return self._config.weight

    def type(self):
        return self._config.type

    def update_bullet_position(self, pybullet, **kwargs):
        pass

    def evaluate(self, **kwargs) -> list:
        return [
            self.position(**kwargs),
            self.velocity(**kwargs),
            self.acceleration(**kwargs)
        ]

    @abstractmethod
    def position(self, **kwargs):
        pass

    @abstractmethod
    def velocity(self, **kwargs):
        pass

    @abstractmethod
    def acceleration(self, **kwargs):
        pass

    @abstractmethod
    def shuffle(self):
        pass



from abc import ABC, abstractmethod
import casadi as ca

t = ca.SX.sym("t", 1)


class ReferenceTrajectory(ABC):
    def __init__(self, n: int, **kwargs):
        assert isinstance(n, int)
        self._n = n

    def n(self):
        return self._n

    @abstractmethod
    def concretize(self):
        pass

    @abstractmethod
    def evaluate(self, t):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['MotionPlanningEnv', 'MotionPlanningGoal', 'MotionPlanningSceneHelpers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'casadi>=3.5.4,<4.0.0,!=3.5.5.post1,!=3.5.5.post2',
 'geomdl>=5.3.1,<6.0.0',
 'matplotlib>=3.3.0,<4.0.0',
 'numpy>=1.19.0,<2.0.0',
 'omegaconf>=2.2.2,<3.0.0',
 'pyquaternion>=0.9.9,<0.10.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.8,<0.9'],
 'bullet': ['pybullet>=3.2.3,<4.0.0']}

setup_kwargs = {
    'name': 'motion-planning-scenes',
    'version': '0.3.5',
    'description': 'Generic motion planning scenes, including goals and obstacles.',
    'long_description': 'None',
    'author': 'Max',
    'author_email': 'm.spahn@tudelft.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)



import numpy as np

from geomdl import BSpline
from geomdl import utilities
import logging

from MotionPlanningSceneHelpers.referenceTrajectory import ReferenceTrajectory


class TrajectoryComponentMissingError(Exception):
    pass


class SplineTrajectory(ReferenceTrajectory):
    def __init__(self, n: int, **kwargs):
        super().__init__(n)
        if "traj" not in kwargs:
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component: traj"
            )
        self._trajDict = kwargs.get("traj")
        if "degree" not in self._trajDict.keys():
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component in trajectory: degree"
            )
        if "controlPoints" not in self._trajDict.keys():
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component in trajectory: controlPoints"
            )
        if "duration" not in self._trajDict.keys():
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component in trajectory: duration"
            )
        self.initialize_spline()

    def initialize_spline(self):
        self._traj = BSpline.Curve()
        self._traj.degree = self._trajDict["degree"]
        list_ctrlpts = [list(val) for val in self._trajDict["controlPoints"]]
        self._traj.ctrlpts = list_ctrlpts
        self._traj.knotvector = utilities.generate_knot_vector(
            self._traj.degree, len(self._traj.ctrlpts)
        )
        self._duration = self._trajDict["duration"]

    def concretize(self):
        pass

    def shuffle(self):
        limit_low = np.array(self._trajDict['low']['controlPoints'])
        limit_high = np.array(self._trajDict['high']['controlPoints'])
        self._trajDict['controlPoints'] = np.random.uniform(limit_low, limit_high, limit_low.shape).tolist()
        self.initialize_spline()

    def timeReparameterize(self, t):
        if t > self._duration:
            return 1.0
        return 0.5 * (1 - np.cos(np.pi * t / self._duration))

    def getScaledDerivatives(self, t):
        t_ref = self.timeReparameterize(t)
        xds = self._traj.derivatives(t_ref, order=2)
        x = np.array(xds[0])
        v_raw = np.array(xds[1])
        a_raw = np.array(xds[2])
        v_scaling = 1 * np.pi / self._duration * np.sin(t * np.pi / self._duration)
        a_scaling = 1 * np.pi / self._duration * np.cos(t * np.pi / self._duration)
        v = v_scaling * v_raw / np.linalg.norm(v_raw)
        if np.linalg.norm(a_raw) < 1e-5:
            logging.warn(f"Assuming zero acceleration on spline, because of large magnitude {np.linalg.norm(a_raw)}")
            a = a_raw * 0
        else:
            a = a_scaling * a_raw / np.linalg.norm(a_raw)
        if t_ref == 1.0:
            v = v_raw * 0.0
            a = a_raw * 0.0
        return x, v, a

    def evaluate(self, t):
        x, v, a = self.getScaledDerivatives(t)
        return [x, v, a]

    def render(self):
        # Import Matplotlib visualization module
        from geomdl.visualization import VisMPL

        # Set the visualization component of the curve
        self._traj.vis = VisMPL.VisCurve3D()

        # Plot the curve
        self._traj.render()
        # TODO: The programm is basically stopped here <09-12-21, mspahn> #
